#pragma once

#ifndef SAW_UNIX
#error "Don't include this"
#endif

#include <csignal>
#include <sys/signalfd.h>

#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>

#include <cassert>
#include <cstring>

#include <errno.h>
#include <unistd.h>

#include <queue>
#include <unordered_map>
#include <vector>

#include "forstio/io.h"

namespace saw {
namespace unix {
constexpr int max_epoll_events = 256;

class UnixEventPort;
class i_fd_owner {
protected:
	UnixEventPort &event_port;

private:
	int m_file_descriptor;
	int m_fd_flags;
	uint32_t m_event_mask;

public:
	i_fd_owner(UnixEventPort &event_port, int file_descriptor, int fd_flags,
			   uint32_t event_mask);

	virtual ~i_fd_owner();

	virtual void notify(uint32_t mask) = 0;

	int fd() const { return m_file_descriptor; }
};

class UnixEventPort final : public EventPort {
private:
	int m_epoll_fd;
	int m_signal_fd;

	sigset_t m_signal_fd_set;

	std::unordered_multimap<Signal, Own<ConveyorFeeder<void>>> m_signal_conveyors;

	int m_pipefds[2];

	std::vector<int> to_unix_signal(Signal signal) const {
		switch (signal) {
		case Signal::User1:
			return {SIGUSR1};
		case Signal::Terminate:
		default:
			return {SIGTERM, SIGQUIT, SIGINT};
		}
	}

	Signal from_unix_signal(int signal) const {
		switch (signal) {
		case SIGUSR1:
			return Signal::User1;
		case SIGTERM:
		case SIGINT:
		case SIGQUIT:
		default:
			return Signal::Terminate;
		}
	}

	void notify_signal_listener(int sig) {
		Signal signal = from_unix_signal(sig);

		auto equal_range = signal_conveyors.equal_range(signal);
		for (auto iter = equal_range.first; iter != equal_range.second;
			 ++iter) {

			if (iter->second) {
				if (iter->second->space() > 0) {
					iter->second->feed();
				}
			}
		}
	}

	bool poll_impl(int time) {
		epoll_event events[max_epoll_events];
		int nfds = 0;
		do {
			nfds = epoll_wait(m_epoll_fd, events, max_epoll_events, time);

			if (nfds < 0) {
				/// @todo error_handling
				return false;
			}

			for (int i = 0; i < nfds; ++i) {
				if (events[i].data.u64 == 0) {
					while (1) {
						struct ::signalfd_siginfo siginfo;
						ssize_t n =
							::read(m_signal_fd, &siginfo, sizeof(siginfo));
						if (n < 0) {
							break;
						}
						assert(n == sizeof(siginfo));

						notify_signal_listener(siginfo.ssi_signo);
					}
				} else if (events[i].data.u64 == 1) {
					uint8_t i;
					if (m_pipefds[0] < 0) {
						continue;
					}
					while (1) {
						ssize_t n = ::recv(m_pipefds[0], &i, sizeof(i), 0);
						if (n < 0) {
							break;
						}
					}
				} else {
					i_fd_owner *owner =
						reinterpret_cast<i_fd_owner *>(events[i].data.ptr);
					if (owner) {
						owner->notify(events[i].events);
					}
				}
			}
		} while (nfds == max_epoll_events);

		return true;
	}

public:
	UnixEventPort() : m_epoll_fd{-1}, m_signal_fd{-1} {
		::signal(SIGPIPE, SIG_IGN);

		m_epoll_fd = ::epoll_create1(EPOLL_CLOEXEC);
		if (m_epoll_fd < 0) {
			return;
		}

		::sigemptyset(&m_signal_fd_set);
		m_signal_fd = ::signalfd(-1, &m_signal_fd_set, SFD_NONBLOCK | SFD_CLOEXEC);
		if (m_signal_fd < 0) {
			return;
		}

		struct epoll_event event;
		memset(&event, 0, sizeof(event));
		event.events = EPOLLIN;
		event.data.u64 = 0;
		::epoll_ctl(m_epoll_fd, EPOLL_CTL_ADD, m_signal_fd, &event);

		int rc = ::pipe2(m_pipefds, O_NONBLOCK | O_CLOEXEC);
		if (rc < 0) {
			return;
		}
		memset(&event, 0, sizeof(event));
		event.events = EPOLLIN;
		event.data.u64 = 1;
		::epoll_ctl(m_epoll_fd, EPOLL_CTL_ADD, m_pipefds[0], &event);
	}

	~UnixEventPort() {
		::close(m_epoll_fd);
		::close(m_signal_fd);
		::close(m_pipefds[0]);
		::close(m_pipefds[1]);
	}

	Conveyor<void> on_signal(Signal signal) override {
		auto caf = newConveyorAndFeeder<void>();

		signal_conveyors.insert(std::make_pair(signal, std::move(caf.feeder)));

		std::vector<int> sig = to_unix_signal(signal);

		for (auto iter = sig.begin(); iter != sig.end(); ++iter) {
			::sigaddset(&signal_fd_set, *iter);
		}
		::sigprocmask(SIG_BLOCK, &m_signal_fd_set, nullptr);
		::signalfd(m_signal_fd, &m_signal_fd_set, SFD_NONBLOCK | SFD_CLOEXEC);

		auto node = Conveyor<void>::fromConveyor(std::move(caf.conveyor));
		return Conveyor<void>::toConveyor(std::move(node));
	}

	void poll() override { poll_impl(0); }

	void wait() override { poll_impl(-1); }

	void wait(const std::chrono::steady_clock::duration &duration) override {
		poll_impl(
			std::chrono::duration_cast<std::chrono::milliseconds>(duration)
				.count());
	}
	void
	wait(const std::chrono::steady_clock::time_point &time_point) override {
		auto now = std::chrono::steady_clock::now();
		if (time_point <= now) {
			poll();
		} else {
			poll_impl(std::chrono::duration_cast<std::chrono::milliseconds>(
						  time_point - now)
						  .count());
		}
	}

	void wake() override {
		/// @todo pipe() in the beginning and write something minor into it like
		/// uint8_t or sth the value itself doesn't matter
		if (m_pipefds[1] < 0) {
			return;
		}
		uint8_t i = 0;
		::send(m_pipefds[1], &i, sizeof(i), MSG_DONTWAIT);
	}

	void subscribe(i_fd_owner &owner, int fd, uint32_t event_mask) {
		if (m_epoll_fd < 0 || fd < 0) {
			return;
		}
		::epoll_event event;
		memset(&event, 0, sizeof(event));
		event.events = event_mask | EPOLLET;
		event.data.ptr = &owner;

		if (::epoll_ctl(m_epoll_fd, EPOLL_CTL_ADD, fd, &event) < 0) {
			/// @todo error_handling
			return;
		}
	}

	void unsubscribe(int fd) {
		if (m_epoll_fd < 0 || fd < 0) {
			return;
		}
		if (::epoll_ctl(m_epoll_fd, EPOLL_CTL_DEL, fd, nullptr) < 0) {
			/// @todo error_handling
			return;
		}
	}
};

ssize_t unix_read(int fd, void *buffer, size_t length);
ssize_t unix_write(int fd, const void *buffer, size_t length);

class unix_io_stream final : public IoStream, public i_fd_owner {
private:
	Own<ConveyorFeeder<void>> m_read_ready = nullptr;
	Own<ConveyorFeeder<void>> m_on_read_disconnect = nullptr;
	Own<ConveyorFeeder<void>> m_write_ready = nullptr;

public:
	unix_io_stream(UnixEventPort &event_port, int file_descriptor, int fd_flags,
				   uint32_t event_mask);

	ErrorOr<size_t> read(void *buffer, size_t length) override;

	Conveyor<void> read_ready() override;

	Conveyor<void> on_read_disconnected() override;

	ErrorOr<size_t> write(const void *buffer, size_t length) override;

	Conveyor<void> write_ready() override;

	/*
		void read(void *buffer, size_t min_length, size_t max_length) override;
		Conveyor<size_t> readDone() override;
		Conveyor<void> readReady() override;

		Conveyor<void> onReadDisconnected() override;

		void write(const void *buffer, size_t length) override;
		Conveyor<size_t> writeDone() override;
		Conveyor<void> writeReady() override;
	*/

	void notify(uint32_t mask) override;
};

class unix_server final : public Server, public i_fd_owner {
private:
	Own<ConveyorFeeder<Own<IoStream>>> m_accept_feeder = nullptr;

public:
	unix_server(UnixEventPort &event_port, int file_descriptor, int fd_flags);

	Conveyor<Own<IoStream>> accept() override;

	void notify(uint32_t mask) override;
};

class unix_datagram final : public Datagram, public i_fd_owner {
private:
	Own<ConveyorFeeder<void>> m_read_ready = nullptr;
	Own<ConveyorFeeder<void>> m_write_ready = nullptr;

public:
	unix_datagram(UnixEventPort &event_port, int file_descriptor, int fd_flags);

	ErrorOr<size_t> read(void *buffer, size_t length) override;
	Conveyor<void> read_ready() override;

	ErrorOr<size_t> write(const void *buffer, size_t length,
						  NetworkAddress &dest) override;
	Conveyor<void> write_ready() override;

	void notify(uint32_t mask) override;
};

/**
 * Helper class which provides potential addresses to NetworkAddress
 */
class socket_address {
private:
	union {
		struct sockaddr generic;
		struct sockaddr_un unix;
		struct sockaddr_in inet;
		struct sockaddr_in6 inet6;
		struct sockaddr_storage storage;
	} m_address;

	socklen_t m_address_length;
	bool m_wildcard;

	socket_address() : m_wildcard{false} {}

public:
	socket_address(const void *sockaddr, socklen_t len, bool wildcard)
		: m_address_length{len}, m_wildcard{wildcard} {
		assert(len <= sizeof(m_address));
		memcpy(&m_address.generic, sockaddr, len);
	}

	int socket(int type) const {
		type |= SOCK_NONBLOCK | SOCK_CLOEXEC;

		int result = ::socket(m_address.generic.sa_family, type, 0);
		return result;
	}

	bool bind(int fd) const {
		if (m_wildcard) {
			int value = 0;
			::setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &value, sizeof(value));
		}
		int error = ::bind(fd, &m_address.generic, m_address_length);
		return error < 0;
	}

	struct ::sockaddr *get_raw() {
		return &m_address.generic;
	}

	const struct ::sockaddr *get_raw() const { return &m_address.generic; }

	socklen_t get_raw_length() const { return m_address_length; }

	static std::vector<socket_address> resolve(std::string_view str,
											   uint16_t port_hint) {
		std::vector<socket_address> results;

		struct ::addrinfo *head;
		struct ::addrinfo hints;
		memset(&hints, 0, sizeof(hints));
		hints.ai_family = AF_UNSPEC;

		std::string port_string = std::to_string(port_hint);
		bool wildcard = str == "*" || str == "::";
		std::string address_string{str};

		int error = ::getaddrinfo(address_string.c_str(), port_string.c_str(),
								  &hints, &head);

		if (error) {
			return {};
		}

		for (struct ::addrinfo *it = head; it != nullptr; it = it->ai_next) {
			if (it->ai_addrlen > sizeof(socket_address::m_address)) {
				continue;
			}
			results.push_back({it->ai_addr, it->ai_addrlen, wildcard});
		}
		::freeaddrinfo(head);
		return results;
	}
};

class unix_network_address final : public OsNetworkAddress {
private:
	const std::string m_path;
	uint16_t m_port_hint;
	std::vector<socket_address> m_addresses;

public:
	unix_network_address(const std::string &path, uint16_t port_hint,
						 std::vector<socket_address> &&addr)
		: m_path{path}, m_port_hint{port_hint}, addresses{std::move(addr)} {}

	const std::string &address() const override;

	uint16_t port() const override;

	// Custom address info
	socket_address &unix_address(size_t i = 0);
	size_t unix_address_size() const;
};

class unix_network final : public Network {
private:
	UnixEventPort &m_event_port;

public:
	unix_network(UnixEventPort &event_port);

	Conveyor<Own<NetworkAddress>>
	resolve_address(const std::string &address,
					uint16_t port_hint = 0) override;

	Own<Server> listen(NetworkAddress &addr) override;

	Conveyor<Own<IoStream>> connect(NetworkAddress &addr) override;

	Own<Datagram> datagram(NetworkAddress &addr) override;
};

class unix_io_provider final : public IoProvider {
private:
	UnixEventPort &m_event_port;
	EventLoop m_event_loop;

	unix_network m_unix_network;

public:
	unix_io_provider(UnixEventPort &port_ref, Own<EventPort> port);

	Network &network() override;

	Own<InputStream> wrap_input_fd(int fd) override;

	EventLoop &event_loop();
};
} // namespace unix
} // namespace saw
