#pragma once

#include "common.h"
#include "error.h"
#include "timer.h"

#include <functional>
#include <limits>
#include <list>
#include <queue>
#include <type_traits>

namespace saw {
class ConveyorStorage;
class conveyor_node {
public:
	conveyor_node();
	virtual ~conveyor_node() = default;

	/**
	 * Internal method to retrieve results from children
	 */
	virtual void get_result(ErrorOrValue &err_or_val) = 0;

	/**
	 * Swap out child with another one
	 */
	virtual ErrorOr<own<conveyor_node>>
	swap_child(Own<conveyor_node> &&swapee_) = 0;

	/**
	 * Retrieve the next storage node
	 */
	virtual ConveyorStorage *next_storage() = 0;

	/**
	 * Notify that a new parent was attached
	 * Only relevant for the feeding nodes
	 */
	virtual void notify_parent_attached(conveyor_node &){};
};

class conveyor_node_with_child_mixin final {
public:
	Own<conveyor_node> child = nullptr;

	conveyor_node_with_child_mixin(own<conveyor_node> &&child_,
								   conveyor_node &owner_);
	~conveyor_node_with_child_mixin() = default;

	/**
	 * Swap out children and return the child ptr, since the caller is the child
	 * itself. Stack needs to be cleared before the child is destroyed, so the
	 * swapped out node is returned as well.
	 */
	ErrorOr<own<conveyor_node>> swap_child(Own<conveyor_node> &&swapee);
};

class conveyor_node_with_parent_mixin final {
public:
	conveyor_node *parent = nullptr;

	ErrorOr<own<conveyor_node>>
	swap_child_of_parent(Own<conveyor_node> &&swapee) {
		SAW_ASSERT(parent) {
			return critical_error(
				"Can't swap child, because parent doesn't exist");
		}

		return parent->swap_child(std::move(swapee));
	}
	void change_parent(conveyor_node *p) { parent = p; }
};

class EventLoop;
class WaitScope;
/*
 * Event class similar to capn'proto.
 * https://github.com/capnproto/capnproto
 */
class event {
private:
	EventLoop &m_loop;
	event **m_prev = nullptr;
	event *m_next = nullptr;

	friend class EventLoop;

public:
	event();
	event(EventLoop &loop);
	virtual ~event();

	virtual void fire() = 0;

	void arm_next();
	void arm_later();
	void arm_last();
	void disarm();

	bool is_armed() const;
};

class ConveyorStorage {
protected:
	ConveyorStorage *parent = nullptr;

public:
	ConveyorStorage();
	virtual ~ConveyorStorage();

	virtual size_t space() const = 0;
	virtual size_t queued() const = 0;
	virtual void child_has_fired() = 0;
	virtual void parent_has_fired() = 0;

	virtual void set_parent(ConveyorStorage *parent) = 0;
	ConveyorStorage *get_parent() const;
};

class conveyor_event_storage : public ConveyorStorage, public event {
public:
	conveyor_event_storage();
	virtual ~conveyor_event_storage() = default;

	void set_parent(ConveyorStorage *parent) override;
};

class conveyor_base {
protected:
	Own<conveyor_node> node;

public:
	conveyor_base(own<conveyor_node> &&node_p);
	virtual ~conveyor_base() = default;

	conveyor_base(conveyor_base &&) = default;
	conveyor_base &operator=(conveyor_base &&) = default;

	void get(ErrorOrValue &err_or_val);
};

template <typename T> class Conveyor;

template <typename T> Conveyor<T> chained_conveyor_type(T *);

// template <typename T> Conveyor<T> chainedConveyorType(Conveyor<T> *);

template <typename T> T remove_error_or_type(T *);

template <typename T> T remove_error_or_type(ErrorOr<T> *);

template <typename T>
using RemoveErrorOr = decltype(removeErrorOrType((T *)nullptr));

template <typename T>
using ChainedConveyors = decltype(chainedConveyorType((T *)nullptr));

template <typename Func, typename T>
using ConveyorResult = ChainedConveyors<RemoveErrorOr<ReturnType<Func, T>>>;

struct PropagateError {
public:
	error operator()(const error &error) const;
	error operator()(error &&error);
};

class conveyor_sink {
private:
	Own<conveyor_node> m_node;

public:
	conveyor_sink();
	conveyor_sink(own<conveyor_node> &&node);

	conveyor_sink(conveyor_sink &&) = default;
	conveyor_sink &operator=(conveyor_sink &&) = default;
};

template <typename T> class MergeConveyorNodeData;

template <typename T> class merge_conveyor {
private:
	Lent<MergeConveyorNodeData<T>> m_data;

public:
	merge_conveyor() = default;
	merge_conveyor(Lent<MergeConveyorNodeData<T>> d);
	~merge_conveyor();

	void attach(Conveyor<T> conveyor);
};

/**
 * Main interface for async operations.
 */
template <typename T> class Conveyor final : public conveyor_base {
public:
	/**
	 * Construct an immediately fulfilled node
	 */
	Conveyor(FixVoid<T> value);

	/**
	 * Construct an immediately failed node
	 */
	Conveyor(error &&err);

	/**
	 * Construct a conveyor with a child node
	 */
	Conveyor(own<conveyor_node> node_p);

	Conveyor(Conveyor<T> &&) = default;
	Conveyor<T> &operator=(Conveyor<T> &&) = default;

	/**
	 * This method converts values or errors from children
	 */
	template <typename Func, typename ErrorFunc = PropagateError>
	[[nodiscard]] ConveyorResult<Func, T>
	then(Func &&func, ErrorFunc &&error_func = PropagateError());

	/**
	 * This method adds a buffer node in the conveyor chains which acts as a
	 * scheduler interrupt point and collects elements up to the supplied limit.
	 */
	[[nodiscard]] Conveyor<T>
	buffer(size_t limit = std::numeric_limits<size_t>::max());

	/**
	 * This method just takes ownership of any supplied types,
	 * which are destroyed when the chain gets destroyed.
	 * Useful for resource lifetime control.
	 */
	template <typename... Args>
	[[nodiscard]] Conveyor<T> attach(Args &&...args);

	/** @todo implement
	 * This method limits the total amount of passed elements
	 * Be careful where you place this node into the chain.
	 * If you meant to fork it and destroy paths you shouldn't place
	 * an interrupt point between the fork and this limiter
	 */
	[[nodiscard]] Conveyor<T> limit(size_t val = 1);

	/**
	 *
	 */
	[[nodiscard]] std::pair<Conveyor<T>, merge_conveyor<T>> merge();

	/**
	 * Moves the conveyor chain into a thread local storage point which drops
	 * every element. Use sink() if you want to control the lifetime of a
	 * conveyor chain
	 */
	template <typename ErrorFunc = PropagateError>
	void detach(ErrorFunc &&err_func = PropagateError());
	/**
	 * Creates a local sink which drops elements, but lifetime control remains
	 * in your hand.
	 */
	template <typename ErrorFunc = PropagateError>
	[[nodiscard]] conveyor_sink sink(ErrorFunc &&error_func = PropagateError());

	/**
	 * If no sink() or detach() is used you have to take elements out of the
	 * chain yourself.
	 */
	ErrorOr<FixVoid<T>> take();

	/** @todo implement
	 * Specifically pump elements through this chain with the provided
	 * wait_scope
	 */
	void poll(WaitScope &wait_scope);

	// helper
	static Conveyor<T> to_conveyor(Own<conveyor_node> node);

	// helper
	static Own<conveyor_node> from_conveyor(Conveyor<T> conveyor);
};

template <typename Func> ConveyorResult<Func, void> exec_later(Func &&func);

/*
 * Join Conveyors into a single one
 */
template <typename... Args>
Conveyor<std::tuple<Args...>>
join_conveyors(std::tuple<Conveyor<Args>...> &conveyors);

template <typename T> class conveyor_feeder {
public:
	virtual ~conveyor_feeder() = default;

	virtual void feed(T &&data) = 0;
	virtual void fail(error &&err) = 0;

	virtual size_t space() const = 0;
	virtual size_t queued() const = 0;

	virtual error swap(Conveyor<T> &&conveyor) noexcept = 0;
};

template <> class conveyor_feeder<void> {
public:
	virtual ~conveyor_feeder() = default;

	virtual void feed(Void &&value = Void{}) = 0;
	virtual void fail(error &&err) = 0;

	virtual size_t space() const = 0;
	virtual size_t queued() const = 0;

	virtual error swap(Conveyor<Void> &&conveyor) noexcept = 0;
};

template <typename T> struct ConveyorAndFeeder {
	Own<conveyor_feeder<T>> feeder;
	Conveyor<T> conveyor;
};

template <typename T> ConveyorAndFeeder<T> new_conveyor_and_feeder();

template <typename T> ConveyorAndFeeder<T> one_time_conveyor_and_feeder();

enum class Signal : uint8_t { Terminate, User1 };

/**
 * Class which acts as a correspondent between the running framework and outside
 * events which may be signals from the operating system or just other threads.
 * Default EventPorts are supplied by setupAsyncIo() in io.h
 */
class event_port {
public:
	virtual ~event_port() = default;

	virtual Conveyor<void> on_signal(Signal signal) = 0;

	virtual void poll() = 0;
	virtual void wait() = 0;
	virtual void wait(const std::chrono::steady_clock::duration &) = 0;
	virtual void wait(const std::chrono::steady_clock::time_point &) = 0;

	virtual void wake() = 0;
};

class SinkConveyorNode;

class conveyor_sink_set final : public event {
private:
	/*
		class Helper final : public Event {
		private:
			void destroySinkConveyorNode(ConveyorNode& sink);
			void fail(Error&& error);

			std::vector<own<ConveyorNode>> sink_nodes;
			std::queue<ConveyorNode*> delete_nodes;
			std::function<void(Error&& error)> error_handler;

		public:
			ConveyorSinks() = default;
			ConveyorSinks(EventLoop& event_loop);

			void add(Conveyor<void> node);

			void fire() override {}
		};

		gin::own<Helper> helper;
	*/
	friend class SinkConveyorNode;

	void destroy_sink_conveyor_node(conveyor_node &sink_node);
	void fail(error &&err);

	std::list<own<conveyor_node>> m_sink_nodes;

	std::queue<conveyor_node *> m_delete_nodes;

	std::function<void(Error &&error)> m_error_handler;

public:
	// ConveyorSinks();
	// ConveyorSinks(EventLoop& event_loop);
	conveyor_sink_set() = default;
	conveyor_sink_set(EventLoop &event_loop);

	void add(Conveyor<void> &&node);

	void fire() override;
};

/*
 * EventLoop class similar to capn'proto.
 * https://github.com/capnproto/capnproto
 */
class EventLoop {
private:
	friend class event;
	event *m_head = nullptr;
	event **m_tail = &m_head;
	event **m_next_insert_point = &m_head;
	event **m_later_insert_point = &m_head;

	bool m_is_runnable = false;

	Own<event_port> m_event_port = nullptr;

	Own<conveyor_sink_set> m_daemon_sink = nullptr;

	// functions
	void set_runnable(bool runnable);

	friend class WaitScope;
	void enter_scope();
	void leave_scope();

	bool turn_loop();
	bool turn();

public:
	EventLoop();
	EventLoop(own<event_port> &&port);
	~EventLoop();

	EventLoop(EventLoop &&) = default;
	EventLoop &operator=(EventLoop &&) = default;

	bool wait();
	bool wait(const std::chrono::steady_clock::duration &);
	bool wait(const std::chrono::steady_clock::time_point &);
	bool poll();

	class event_port *event_port();

	conveyor_sink_set &daemon();
};

/*
 * WaitScope class similar to capn'proto.
 * https://github.com/capnproto/capnproto
 */
class WaitScope {
private:
	EventLoop &m_loop;

public:
	WaitScope(EventLoop &loop);
	~WaitScope();

	void wait();
	void wait(const std::chrono::steady_clock::duration &);
	void wait(const std::chrono::steady_clock::time_point &);
	void poll();
};

template <typename Func> ConveyorResult<Func, void> yield_next(Func &&func);

template <typename Func> ConveyorResult<Func, void> yield_later(Func &&func);

template <typename Func> ConveyorResult<Func, void> yield_last(Func &&func);
} // namespace saw

// Secret stuff
// Aka private semi hidden classes
namespace saw {

template <typename Out, typename In> struct FixVoidCaller {
	template <typename Func> static Out apply(Func &func, In &&in) {
		return func(std::move(in));
	}
};

template <typename Out> struct FixVoidCaller<Out, Void> {
	template <typename Func> static Out apply(Func &func, Void &&in) {
		(void)in;
		return func();
	}
};

template <typename In> struct FixVoidCaller<Void, In> {
	template <typename Func> static Void apply(Func &func, In &&in) {
		func(std::move(in));
		return Void{};
	}
};

template <> struct FixVoidCaller<Void, Void> {
	template <typename Func> static Void apply(Func &func, Void &&in) {
		(void)in;
		func();
		return Void{};
	}
};

template <typename T> class AdaptConveyorNode;

template <typename T>
class adapt_conveyor_feeder final : public conveyor_feeder<UnfixVoid<T>> {
private:
	AdaptConveyorNode<T> *m_feedee = nullptr;

public:
	~adapt_conveyor_feeder();

	void set_feedee(AdaptConveyorNode<T> *feedee);

	void feed(T &&value) override;
	void fail(error &&err) override;

	size_t space() const override;
	size_t queued() const override;

	error swap(Conveyor<T> &&conv) noexcept override;
};

template <typename T>
class AdaptConveyorNode final : public conveyor_node,
								public conveyor_event_storage {
private:
	adapt_conveyor_feeder<T> *m_feeder = nullptr;

	std::queue<ErrorOr<UnfixVoid<T>>> m_storage;

	conveyor_node_with_parent_mixin m_parent_node;

public:
	AdaptConveyorNode();
	~AdaptConveyorNode();

	void set_feeder(adapt_conveyor_feeder<T> *feeder);

	void feed(T &&value);
	void fail(error &&err);

	// ConveyorNode
	void get_result(ErrorOrValue &err_or_val) override;

	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&swapee) noexcept override;

	ConveyorStorage *next_storage() noexcept override;
	void notify_parent_attached(conveyor_node &) noexcept override;

	// ConveyorStorage
	size_t space() const override;
	size_t queued() const override;

	void child_has_fired() override;
	void parent_has_fired() override;

	// Event
	void fire() override;
};

template <typename T> class OneTimeConveyorNode;

template <typename T>
class one_time_conveyor_feeder final : public conveyor_feeder<UnfixVoid<T>> {
private:
	OneTimeConveyorNode<T> *m_feedee = nullptr;

public:
	~one_time_conveyor_feeder();

	void set_feedee(OneTimeConveyorNode<T> *feedee);

	void feed(T &&value) override;
	void fail(error &&err) override;

	size_t space() const override;
	size_t queued() const override;
};

template <typename T>
class OneTimeConveyorNode final : public conveyor_node,
								  public ConveyorStorage,
								  public event {
private:
	one_time_conveyor_feeder<T> *m_feeder = nullptr;

	bool m_passed = false;
	Maybe<ErrorOr<T>> m_storage = std::nullopt;

public:
	~OneTimeConveyorNode();

	void set_feeder(one_time_conveyor_feeder<T> *feeder);

	void feed(T &&value);
	void fail(error &&err);

	// ConveyorNode
	void get_result(ErrorOrValue &err_or_val) override;

	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&swapee) override;

	// ConveyorStorage
	size_t space() const override;
	size_t queued() const override;

	void child_has_fired() override {}
	void parent_has_fired() override;

	// Event
	void fire() override;
};

/**
 * This class buffers and saves incoming elements and acts as an interrupt node
 * for processing calls
 */
class queue_buffer_conveyor_node_base : public conveyor_node,
										public conveyor_event_storage {
protected:
	conveyor_node_with_child_mixin child_mixin;

public:
	queue_buffer_conveyor_node_base(own<conveyor_node> child_)
		: conveyor_event_storage{}, child_mixin{std::move(child_), *this} {}
	virtual ~queue_buffer_conveyor_node_base() = default;

	/**
	 * Use mixin
	 */
	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&swapee_) noexcept override;

	ConveyorStorage *next_storage() noexcept override {
		return static_cast<ConveyorStorage *>(this);
	}
};

template <typename T>
class queue_buffer_conveyor_node final
	: public queue_buffer_conveyor_node_base {
private:
	std::queue<ErrorOr<T>> m_storage;
	size_t m_max_store;

public:
	queue_buffer_conveyor_node(own<conveyor_node> dep, size_t max_size)
		: queue_buffer_conveyor_node_base{std::move(dep)}, max_store{max_size} {
	}
	// Event
	void fire() override;
	// ConveyorNode
	void get_result(ErrorOrValue &eov) noexcept override;

	// ConveyorStorage
	size_t space() const override;
	size_t queued() const override;

	void child_has_fired() override;
	void parent_has_fired() override;
};

class attach_conveyor_node_base : public conveyor_node {
protected:
	conveyor_node_with_child_mixin child_mixin;

public:
	attach_conveyor_node_base(own<conveyor_node> &&child_)
		: child_mixin{std::move(child_), *this} {}

	virtual ~attach_conveyor_node_base() = default;

	void get_result(ErrorOrValue &err_or_val) noexcept override;

	/**
	 * Use mixin
	 */
	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&swapee_) noexcept override;

	ConveyorStorage *next_storage() noexcept override;
};

template <typename... Args>
class attach_conveyor_node final : public attach_conveyor_node_base {
public:
	attach_conveyor_node(own<conveyor_node> &&dep, Args &&...args)
		: attach_conveyor_node_base(std::move(dep)), attached_data{
														 std::move(args...)} {}

private:
	std::tuple<Args...> attached_data;
};

class convert_conveyor_node_base : public conveyor_node {
public:
	convert_conveyor_node_base(own<conveyor_node> &&dep);
	virtual ~convert_conveyor_node_base() = default;

	void get_result(ErrorOrValue &err_or_val) override;

	virtual void get_impl(ErrorOrValue &err_or_val) = 0;

	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&swapee) noexcept override;

	ConveyorStorage *next_storage() noexcept override;

protected:
	conveyor_node_with_child_mixin child_mixin;
};

template <typename T, typename DepT, typename Func, typename ErrorFunc>
class convert_conveyor_node final : public convert_conveyor_node_base {
private:
	Func m_func;
	ErrorFunc m_error_func;

	static_assert(std::is_same<DepT, RemoveErrorOr<DepT>>::value,
				  "Should never be of type ErrorOr");

public:
	convert_conveyor_node(own<conveyor_node> &&dep, Func &&func,
						  ErrorFunc &&error_func_)
		: convert_conveyor_node_base(std::move(dep)), func{std::move(func)},
		  error_func{std::move(error_func:)} {}

	void get_impl(ErrorOrValue &err_or_val) noexcept override {
		ErrorOr<UnfixVoid<DepT>> dep_eov;
		ErrorOr<UnfixVoid<RemoveErrorOr<T>>> &eov =
			err_or_val.as<UnfixVoid<RemoveErrorOr<T>>>();
		if (child_mixin.child) {
			child_mixin.child->get_result(dep_eov);
			if (dep_eov.isValue()) {
				// NOLINTBEGIN
				try {

					eov = FixVoidCaller<T, DepT>::apply(
						func, std::move(dep_eov.value()));
				} catch (const std::bad_alloc &) {
					eov = criticalError("Out of memory");
				} catch (const std::exception &) {
					eov = criticalError(
						"Exception in chain occured. Return ErrorOr<T> if you "
						"want to handle errors which are recoverable");
				}
				// NOLINTEND
			} else if (dep_eov.isError()) {
				eov = error_func(std::move(dep_eov.error()));
			} else {
				eov = criticalError("No value set in dependency");
			}
		} else {
			eov = criticalError("Conveyor doesn't have child");
		}
	}
};

class SinkConveyorNode final : public conveyor_node,
							   public conveyor_event_storage {
private:
	conveyor_node_with_child_mixin m_child_mixin;
	conveyor_sink_set *m_conveyor_sink;

public:
	SinkConveyorNode(own<conveyor_node> node, conveyor_sink_set &conv_sink)
		: conveyor_event_storage{}, child_mixin{std::move(node), *this},
		  conveyor_sink{&conv_sink} {}

	SinkConveyorNode(own<conveyor_node> node)
		: conveyor_event_storage{}, child_mixin{std::move(node), *this},
		  conveyor_sink{nullptr} {}

	// Event only queued if a critical error occured
	void fire() override {
		// Queued for destruction of children, because this acts as a sink and
		// no other event should be here
		child_mixin.child = nullptr;

		if (conveyor_sink) {
			conveyor_sink->destroy_sink_conveyor_node(*this);
			conveyor_sink = nullptr;
		}
	}

	// ConveyorStorage
	size_t space() const override { return 1; }
	size_t queued() const override { return 0; }

	// ConveyorNode
	void get_result(ErrorOrValue &err_or_val) noexcept override {
		err_or_val.as<Void>() =
			criticalError("In a sink node no result can be returned");
	}

	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&swapee) noexcept override {
		return child_mixin.swap_child(std::move(swapee));
	}

	// ConveyorStorage
	void child_has_fired() override {
		if (child_mixin.child) {
			ErrorOr<void> dep_eov;
			child_mixin.child->get_result(dep_eov);
			if (dep_eov.isError()) {
				if (dep_eov.error().isCritical()) {
					if (!is_armed()) {
						arm_last();
					}
				}
				if (conveyor_sink) {
					conveyor_sink->fail(std::move(dep_eov.error()));
				}
			}
		}
	}

	/*
	 * No parent needs to be fired since we always have space
	 */
	void parent_has_fired() override {}

	ConveyorStorage *next_storage() override {
		// Should never happen though
		assert(false);
		return nullptr;
		// return static_cast<ConveyorStorage*>(this);
	}
};

class immediate_conveyor_node_base : public conveyor_node,
									 public conveyor_event_storage {
private:
public:
	immediate_conveyor_node_base();

	virtual ~immediate_conveyor_node_base() = default;

	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&swapee) noexcept override {
		(void)swapee;
		return recoverableError("Node doesn't support swapping");
	}

	ConveyorStorage *next_storage() noexcept override {
		return static_cast<ConveyorStorage *>(this);
	}
};

template <typename T>
class immediate_conveyor_node final : public immediate_conveyor_node_base {
private:
	ErrorOr<FixVoid<T>> m_value;
	uint8_t m_retrieved;

public:
	immediate_conveyor_node(FixVoid<T> &&val);
	immediate_conveyor_node(Error &&err);

	// ConveyorStorage
	size_t space() const override;
	size_t queued() const override;

	void child_has_fired() override;
	void parent_has_fired() override;

	// ConveyorNode
	void get_result(ErrorOrValue &err_or_val) noexcept override {
		if (m_retrieved > 0) {
			err_or_val.as<FixVoid<T>>() =
				makeError("Already taken value", Error::Code::Exhausted);
		} else {
			err_or_val.as<FixVoid<T>>() = std::move(value);
		}
		if (queued() > 0) {
			++m_retrieved;
		}
	}

	// Event
	void fire() override;
};

/*
 * Collects every incoming value and throws it in one lane
 */
class merge_conveyor_node_base : public conveyor_node,
								 public conveyor_event_storage {
public:
	merge_conveyor_node_base();

	virtual ~merge_conveyor_node_base() = default;

	ConveyorStorage *next_storage() noexcept override {
		return static_cast<ConveyorStorage *>(this);
	}
};

template <typename T>
class merge_conveyor_node : public merge_conveyor_node_base {
private:
	class appendage final : public conveyor_node, public ConveyorStorage {
	public:
		own<conveyor_node> child;
		merge_conveyor_node *merger;

		Maybe<ErrorOr<FixVoid<T>>> error_or_value_data;

	public:
		appendage(own<conveyor_node> n, merge_conveyor_node &m)
			: ConveyorStorage{}, child{std::move(n)}, merger{&m},
			  error_or_value_data{std::nullopt} {}

		bool child_storage_has_element_queued() const {
			if (!child) {
				return false;
			}
			ConveyorStorage *storage = child->next_storage();
			if (storage) {
				return storage->queued() > 0;
			}
			return false;
		}

		void get_appendage_result(ErrorOrValue &eov);

		/**
		 * ConveyorNode
		 */
		ErrorOr<own<conveyor_node>>
		swap_child(own<conveyor_node> &&swapee_) override;

		ConveyorStorage *next_storage() noexcept override {
			return static_cast<ConveyorStorage *>(this);
		}

		void get_result(ErrorOrValue &err_or_val) override;

		/**
		 * ConveyorStorage
		 */
		size_t space() const override;

		size_t queued() const override;

		void child_has_fired() override;

		void parent_has_fired() override;

		void set_parent(ConveyorStorage *par) override;
	};

	friend class MergeConveyorNodeData<T>;
	friend class appendage;

	Our<MergeConveyorNodeData<T>> m_data;
	size_t m_next_appendage = 0;

public:
	merge_conveyor_node(Our<MergeConveyorNodeData<T>> data);
	~merge_conveyor_node();
	// ConveyorNode
	ErrorOr<own<conveyor_node>>
	swap_child(own<conveyor_node> &&c) noexcept override;

	// Event
	void get_result(ErrorOrValue &err_or_val) noexcept override;

	void fire() override;

	// ConveyorStorage
	size_t space() const override;
	size_t queued() const override;
	void child_has_fired() override;
	void parent_has_fired() override;
};

template <typename T> class MergeConveyorNodeData {
public:
	std::vector<own<typename merge_conveyor_node<T>::Appendage>> appendages;

	merge_conveyor_node<T> *merger = nullptr;

public:
	void attach(Conveyor<T> conv);

	void governing_node_destroyed();
};

/*
class JoinConveyorNodeBase : public ConveyorNode, public ConveyorEventStorage {
private:

public:
};

template <typename... Args>
class JoinConveyorNode final : public JoinConveyorNodeBase {
private:
	template<typename T>
	class Appendage : public ConveyorEventStorage {
	private:
		Maybe<T> data = std::nullopt;

	public:
		size_t space() const override;
		size_t queued() const override;

		void fire() override;
		void getResult(ErrorOrValue& eov) override;
	};

	std::tuple<Appendage<Args>...> appendages;

public:
};

*/

} // namespace saw

#include "async.tmpl.h"
