#pragma once

#include "common.h"
#include "error.h"

#include <cassert>
// Template inlining

#include <iostream>

namespace saw {

template <typename Func> ConveyorResult<Func, void> exec_later(Func &&func) {
	Conveyor<void> conveyor{FixVoid<void>{}};
	return conveyor.then(std::move(func));
}

template <typename T>
Conveyor<T>::Conveyor(FixVoid<T> value) : conveyor_base(nullptr) {
	// Is there any way to do  this?
	// @todo new ConveyorBase constructor for Immediate values

	Own<immediate_conveyor_node<FixVoid<T>>> immediate =
		heap<immediate_conveyor_node<FixVoid<T>>>(std::move(value));

	if (!immediate) {
		return;
	}

	node = std::move(immediate);
}

template <typename T>
Conveyor<T>::Conveyor(Error &&error) : conveyor_base(nullptr) {
	Own<immediate_conveyor_node<FixVoid<T>>> immediate =
		heap<immediate_conveyor_node<FixVoid<T>>>(std::move(error));

	if (!immediate) {
		return;
	}

	node = std::move(immediate);
}

template <typename T>
Conveyor<T>::Conveyor(Own<conveyor_node> node_p)
	: conveyor_base{std::move(node_p)} {}

template <typename T>
template <typename Func, typename ErrorFunc>
ConveyorResult<Func, T> Conveyor<T>::then(Func &&func, ErrorFunc &&error_func) {
	Own<conveyor_node> conversion_node =
		heap<convert_conveyor_node<FixVoid<ReturnType<Func, T>>, FixVoid<T>,
								   Func, ErrorFunc>>(
			std::move(node), std::move(func), std::move(error_func));

	return Conveyor<RemoveErrorOr<ReturnType<Func, T>>>::to_conveyor(
		std::move(conversion_node));
}

template <typename T> Conveyor<T> Conveyor<T>::buffer(size_t size) {
	SAW_ASSERT(node) { return Conveyor<T>{Own<conveyor_node>{nullptr}}; }
	ConveyorStorage *storage = node->next_storage();
	SAW_ASSERT(storage) { return Conveyor<T>{Own<conveyor_node>{nullptr}}; }

	Own<queue_buffer_conveyor_node<FixVoid<T>>> storage_node =
		heap<queue_buffer_conveyor_node<FixVoid<T>>>(std::move(node), size);

	ConveyorStorage *storage_ptr =
		static_cast<ConveyorStorage *>(storage_node.get());

	storage->set_parent(storage_ptr);
	return Conveyor<T>{std::move(storage_node)};
}

template <typename T>
template <typename... Args>
Conveyor<T> Conveyor<T>::attach(Args &&...args) {
	Own<attach_conveyor_node<Args...>> attach_node =
		heap<attach_conveyor_node<Args...>>(std::move(node),
											std::move(args...));
	return Conveyor<T>{std::move(attach_node)};
}

template <typename T>
std::pair<Conveyor<T>, merge_conveyor<T>> Conveyor<T>::merge() {
	Our<MergeConveyorNodeData<T>> data = share<MergeConveyorNodeData<T>>();

	Own<merge_conveyor_node<T>> merge_node = heap<merge_conveyor_node<T>>(data);

	SAW_ASSERT(node) {
		return std::make_pair(Conveyor<T>{Own<conveyor_node>{nullptr}},
							  merge_conveyor<T>{});
	}
	ConveyorStorage *storage = node->next_storage();
	SAW_ASSERT(storage) {
		return std::make_pair(Conveyor<T>{Own<conveyor_node>{nullptr}},
							  merge_conveyor<T>{});
	}

	data->attach(Conveyor<T>::to_conveyor(std::move(node)));

	merge_conveyor<T> node_ref{data};

	return std::make_pair(Conveyor<T>{std::move(merge_node)},
						  std::move(node_ref));
}

template <>
template <typename ErrorFunc>
conveyor_sink Conveyor<void>::sink(ErrorFunc &&error_func) {
	ConveyorStorage *storage = node->next_storage();
	SAW_ASSERT(storage) { return conveyor_sink{}; }

	Own<SinkConveyorNode> sink_node = heap<SinkConveyorNode>(std::move(node));
	ConveyorStorage *storage_ptr =
		static_cast<ConveyorStorage *>(sink_node.get());

	storage->set_parent(storage_ptr);

	return conveyor_sink{std::move(sink_node)};
}

void detachConveyor(Conveyor<void> &&conveyor);

template <typename T>
template <typename ErrorFunc>
void Conveyor<T>::detach(ErrorFunc &&func) {
	detachConveyor(std::move(then([](T &&) {}, std::move(func))));
}

template <>
template <typename ErrorFunc>
void Conveyor<void>::detach(ErrorFunc &&func) {
	detachConveyor(std::move(then([]() {}, std::move(func))));
}

template <typename T>
Conveyor<T> Conveyor<T>::to_conveyor(Own<conveyor_node> node) {
	return Conveyor<T>{std::move(node)};
}

template <typename T>
Own<conveyor_node> Conveyor<T>::from_conveyor(Conveyor<T> conveyor) {
	return std::move(conveyor.node);
}

template <typename T> ErrorOr<FixVoid<T>> Conveyor<T>::take() {
	SAW_ASSERT(node) {
		return ErrorOr<FixVoid<T>>{criticalError("Conveyor in invalid state")};
	}
	ConveyorStorage *storage = node->next_storage();
	if (storage) {
		if (storage->queued() > 0) {
			ErrorOr<FixVoid<T>> result;
			node->get_result(result);
			return result;
		} else {
			return ErrorOr<FixVoid<T>>{
				recoverableError("Conveyor buffer has no elements")};
		}
	} else {
		return ErrorOr<FixVoid<T>>{
			criticalError("Conveyor node has no child storage")};
	}
}

template <typename T> ConveyorAndFeeder<T> new_conveyor_and_feeder() {
	Own<adapt_conveyor_feeder<FixVoid<T>>> feeder =
		heap<adapt_conveyor_feeder<FixVoid<T>>>();
	Own<AdaptConveyorNode<FixVoid<T>>> node =
		heap<AdaptConveyorNode<FixVoid<T>>>();

	feeder->setFeedee(node.get());
	node->setFeeder(feeder.get());

	return ConveyorAndFeeder<T>{std::move(feeder),
								Conveyor<T>::toConveyor(std::move(node))};
}

// QueueBuffer
template <typename T> void queue_buffer_conveyor_node<T>::fire() {
	if (child_mixin.child) {
		if (!storage.empty()) {
			if (storage.front().isError()) {
				if (storage.front().error().isCritical()) {
					child_mixin.child = nullptr;
				}
			}
		}
	}

	bool has_space_before_fire = space() > 0;

	if (parent) {
		parent->child_has_fired();
		if (!storage.empty() && parent->space() > 0) {
			arm_later();
		}
	}

	if (!child_mixin.child) {
		while (!storage.empty()) {
			storage.pop();
		}
		return;
	}

	ConveyorStorage *ch_storage = child_mixin.child->next_storage();
	if (ch_storage && !has_space_before_fire) {
		ch_storage->parent_has_fired();
	}
}

template <typename T>
void queue_buffer_conveyor_node<T>::get_result(ErrorOrValue &eov) noexcept {
	ErrorOr<T> &err_or_val = eov.as<T>();
	err_or_val = std::move(storage.front());
	storage.pop();
}

template <typename T> size_t queue_buffer_conveyor_node<T>::space() const {
	return max_store - storage.size();
}

template <typename T> size_t queue_buffer_conveyor_node<T>::queued() const {
	return storage.size();
}

template <typename T> void queue_buffer_conveyor_node<T>::child_has_fired() {
	if (child_mixin.child && storage.size() < max_store) {
		ErrorOr<T> eov;
		child_mixin.child->get_result(eov);

		if (eov.isError()) {
			if (eov.error().isCritical()) {
			}
		}

		storage.push(std::move(eov));
		if (!is_armed()) {
			arm_later();
		}
	}
}

template <typename T> void queue_buffer_conveyor_node<T>::parent_has_fired() {
	SAW_ASSERT(parent) { return; }

	if (parent->space() == 0) {
		return;
	}

	if (queued() > 0) {
		arm_later();
	}
}

template <typename T>
immediate_conveyor_node<T>::immediate_conveyor_node(FixVoid<T> &&val)
	: value{std::move(val)}, retrieved{0} {}

template <typename T>
immediate_conveyor_node<T>::immediate_conveyor_node(Error &&error)
	: value{std::move(error)}, retrieved{0} {}

template <typename T> size_t immediate_conveyor_node<T>::space() const {
	return 0;
}

template <typename T> size_t immediate_conveyor_node<T>::queued() const {
	return retrieved > 1 ? 0 : 1;
}

template <typename T> void immediate_conveyor_node<T>::child_has_fired() {
	// Impossible case
	assert(false);
}

template <typename T> void immediate_conveyor_node<T>::parent_has_fired() {
	SAW_ASSERT(parent) { return; }
	assert(parent->space() > 0);

	if (queued() > 0) {
		arm_next();
	}
}

template <typename T> void immediate_conveyor_node<T>::fire() {

	if (parent) {
		parent->child_has_fired();
		if (queued() > 0 && parent->space() > 0) {
			arm_last();
		}
	}
}

template <typename T>
merge_conveyor<T>::merge_conveyor(Lent<MergeConveyorNodeData<T>> d)
	: data{std::move(d)} {}

template <typename T> merge_conveyor<T>::~merge_conveyor() {}

template <typename T> void merge_conveyor<T>::attach(Conveyor<T> conveyor) {
	auto sp = data.lock();
	SAW_ASSERT(sp) { return; }

	sp->attach(std::move(conveyor));
}

template <typename T>
merge_conveyor_node<T>::merge_conveyor_node(Our<MergeConveyorNodeData<T>> d)
	: data{d} {
	SAW_ASSERT(data) { return; }

	data->merger = this;
}

template <typename T> merge_conveyor_node<T>::~merge_conveyor_node() {}

template <typename T>
ErrorOr<Own<conveyor_node>>
merge_conveyor_node<T>::swap_child(Own<conveyor_node> &&swapee_) noexcept {
	(void)swapee_;
	return criticalError(
		"MergeConveyorNode<T>::Appendage should block calls to this class");
}

template <typename T>
void merge_conveyor_node<T>::get_result(ErrorOrValue &eov) noexcept {
	ErrorOr<FixVoid<T>> &err_or_val = eov.as<FixVoid<T>>();

	SAW_ASSERT(data) { return; }

	/// @todo search appendages for result

	auto &appendages = data->appendages;
	next_appendage = std::min(appendages.size(), next_appendage);

	for (size_t i = next_appendage; i < appendages.size(); ++i) {
		if (appendages[i]->queued() > 0) {
			err_or_val = std::move(appendages[i]->error_or_value.value());
			appendages[i]->error_or_value = std::nullopt;
			next_appendage = i + 1;
			return;
		}
	}
	for (size_t i = 0; i < next_appendage; ++i) {
		if (appendages[i]->queued() > 0) {
			err_or_val = std::move(appendages[i]->error_or_value.value());
			appendages[i]->error_or_value = std::nullopt;
			next_appendage = i + 1;
			return;
		}
	}

	err_or_val = criticalError("No value in Merge Appendages");
}

template <typename T> void merge_conveyor_node<T>::fire() {
	SAW_ASSERT(queued() > 0) { return; }

	if (parent) {
		parent->child_has_fired();

		if (queued() > 0 && parent->space() > 0) {
			arm_later();
		}
	}
}

template <typename T> size_t merge_conveyor_node<T>::space() const { return 0; }

template <typename T> size_t merge_conveyor_node<T>::queued() const {
	SAW_ASSERT(data) { return 0; }

	size_t queue_count = 0;

	for (auto &iter : data->appendages) {
		queue_count += iter->queued();
	}

	return queue_count;
}

template <typename T> void merge_conveyor_node<T>::child_has_fired() {
	/// This can never happen
	assert(false);
}

template <typename T> void merge_conveyor_node<T>::parent_has_fired() {
	SAW_ASSERT(parent) { return; }
	if (queued() > 0) {
		if (parent->space() > 0) {
			arm_later();
		}
	}
}

/**
 * MergeConveyorNode<T>::Apendage
 */

template <typename T>
ErrorOr<Own<conveyor_node>>
merge_conveyor_node<T>::appendage::swap_child(Own<conveyor_node> &&swapee_) {
	Own<conveyor_node> old_child = std::move(child);

	child = std::move(swapee_);

	// This case should never happen
	SAW_ASSERT(old_child) { return criticalError("No child exists"); }

	return old_child;
}

template <typename T>
void merge_conveyor_node<T>::appendage::get_result(ErrorOrValue &eov) {
	ErrorOr<FixVoid<T>> &err_or_val = eov.as<FixVoid<T>>();

	SAW_ASSERT(queued() > 0) {
		err_or_val = criticalError("No element queued in Merge Appendage Node");
		return;
	}

	err_or_val = std::move(error_or_value.value());
	error_or_value = std::nullopt;
}

template <typename T> size_t merge_conveyor_node<T>::appendage::space() const {
	SAW_ASSERT(merger) { return 0; }

	if (error_or_value.has_value()) {
		return 0;
	}

	return 1;
}

template <typename T> size_t merge_conveyor_node<T>::appendage::queued() const {
	SAW_ASSERT(merger) { return 0; }

	if (error_or_value.has_value()) {
		return 1;
	}

	return 0;
}

/// @todo delete this function. Replaced by the regular getResult
template <typename T>
void merge_conveyor_node<T>::appendage::get_appendage_result(
	ErrorOrValue &eov) {
	ErrorOr<FixVoid<T>> &err_or_val = eov.as<FixVoid<T>>();

	SAW_ASSERT(queued() > 0) {
		err_or_val = criticalError("No element queued in Merge Appendage Node");
		return;
	}

	err_or_val = std::move(error_or_value.value());
	error_or_value = std::nullopt;
}

template <typename T>
void merge_conveyor_node<T>::appendage::child_has_fired() {
	SAW_ASSERT(!error_or_value.has_value()) { return; }
	ErrorOr<FixVoid<T>> eov;
	child->get_result(eov);

	error_or_value = std::move(eov);

	if (!merger->is_armed()) {
		merger->arm_later();
	}
}

template <typename T>
void merge_conveyor_node<T>::appendage::parent_has_fired() {
	ConveyorStorage *child_storage = child->next_storage();
	if (child_storage) {
		child_storage->parent_has_fired();
	}
}

template <typename T>
void merge_conveyor_node<T>::appendage::set_parent(ConveyorStorage *par) {
	SAW_ASSERT(merger) { return; }

	SAW_ASSERT(child) { return; }

	parent = par;
}

template <typename T>
void MergeConveyorNodeData<T>::attach(Conveyor<T> conveyor) {
	auto nas = Conveyor<T>::fromConveyor(std::move(conveyor));
	SAW_ASSERT(nas) { return; }
	ConveyorStorage *storage = nas->nextStorage();
	SAW_ASSERT(storage) { return; }

	auto merge_node_appendage =
		heap<typename merge_conveyor_node<T>::Appendage>(std::move(nas),
														 *merger);
	auto merge_node_appendage_ptr = merge_node_appendage.get();

	storage->set_parent(merge_node_appendage.get());

	SAW_ASSERT(merger) { return; }

	ConveyorStorage *mrg_storage = merger->nextStorage();
	SAW_ASSERT(mrg_storage) { return; }

	merge_node_appendage->setParent(mrg_storage);

	appendages.push_back(std::move(merge_node_appendage));

	/// @todo return this. necessary? maybe for the weird linking setup
	/// maybe not
	// return merge_node_appendage_ptr;
}

template <typename T>
void MergeConveyorNodeData<T>::governing_node_destroyed() {
	appendages.clear();
	merger = nullptr;
}

template <typename T> adapt_conveyor_feeder<T>::~adapt_conveyor_feeder() {
	if (feedee) {
		feedee->setFeeder(nullptr);
		feedee = nullptr;
	}
}

template <typename T>
void adapt_conveyor_feeder<T>::set_feedee(AdaptConveyorNode<T> *feedee_p) {
	feedee = feedee_p;
}

template <typename T> void adapt_conveyor_feeder<T>::feed(T &&value) {
	if (feedee) {
		feedee->feed(std::move(value));
	}
}

template <typename T> void adapt_conveyor_feeder<T>::fail(Error &&error) {
	if (feedee) {
		feedee->fail(std::move(error));
	}
}

template <typename T> size_t adapt_conveyor_feeder<T>::queued() const {
	if (feedee) {
		return feedee->queued();
	}
	return 0;
}

template <typename T> size_t adapt_conveyor_feeder<T>::space() const {
	if (feedee) {
		return feedee->space();
	}
	return 0;
}

template <typename T>
Error adapt_conveyor_feeder<T>::swap(Conveyor<T> &&conveyor) noexcept {
	SAW_ASSERT(feedee) { return criticalError("No feedee connected"); }

	auto node = Conveyor<T>::fromConveyor(std::move(conveyor));

	feedee->swapChild(std::move(node));

	return noError();
}

template <typename T>
AdaptConveyorNode<T>::AdaptConveyorNode() : conveyor_event_storage{} {}

template <typename T> AdaptConveyorNode<T>::~AdaptConveyorNode() {
	if (feeder) {
		feeder->setFeedee(nullptr);
		feeder = nullptr;
	}
}

template <typename T>
ErrorOr<Own<conveyor_node>>
AdaptConveyorNode<T>::swap_child(Own<conveyor_node> &&swapee) noexcept {
	// This should return the owning pointer of this instance
	auto myself_err = parent_node.swap_child_of_parent(std::move(swapee));

	if (myself_err.isError()) {
		return myself_err;
	}

	auto &myself = myself_err.value();

	assert(myself.get() == this);

	return myself_err;
}

template <typename T>
ConveyorStorage *AdaptConveyorNode<T>::next_storage() noexcept {
	return static_cast<ConveyorStorage *>(this);
}

template <typename T>
void AdaptConveyorNode<T>::notify_parent_attached(conveyor_node &par) noexcept {
	parent_node.change_parent(&par);
}

template <typename T>
void AdaptConveyorNode<T>::set_feeder(adapt_conveyor_feeder<T> *feeder_p) {
	feeder = feeder_p;
}

template <typename T> void AdaptConveyorNode<T>::feed(T &&value) {
	storage.push(std::move(value));
	arm_next();
}

template <typename T> void AdaptConveyorNode<T>::fail(Error &&error) {
	storage.push(std::move(error));
	arm_next();
}

template <typename T> size_t AdaptConveyorNode<T>::queued() const {
	return storage.size();
}

template <typename T> size_t AdaptConveyorNode<T>::space() const {
	return std::numeric_limits<size_t>::max() - storage.size();
}

template <typename T>
void AdaptConveyorNode<T>::get_result(ErrorOrValue &err_or_val) {
	if (!storage.empty()) {
		err_or_val.as<T>() = std::move(storage.front());
		storage.pop();
	} else {
		err_or_val.as<T>() =
			criticalError("Signal for retrieval of storage sent even though no "
						  "data is present");
	}
}

template <typename T> void AdaptConveyorNode<T>::child_has_fired() {
	// Adapt node has no children
	assert(false);
}

template <typename T> void AdaptConveyorNode<T>::parent_has_fired() {
	SAW_ASSERT(parent) { return; }

	if (parent->space() == 0) {
		return;
	}
}

template <typename T> void AdaptConveyorNode<T>::fire() {
	if (parent) {
		parent->child_has_fired();

		if (storage.size() > 0) {
			arm_later();
		}
	}
}

template <typename T> one_time_conveyor_feeder<T>::~one_time_conveyor_feeder() {
	if (feedee) {
		feedee->setFeeder(nullptr);
		feedee = nullptr;
	}
}

template <typename T>
void one_time_conveyor_feeder<T>::set_feedee(OneTimeConveyorNode<T> *feedee_p) {
	feedee = feedee_p;
}

template <typename T> void one_time_conveyor_feeder<T>::feed(T &&value) {
	if (feedee) {
		feedee->feed(std::move(value));
	}
}

template <typename T> void one_time_conveyor_feeder<T>::fail(Error &&error) {
	if (feedee) {
		feedee->fail(std::move(error));
	}
}

template <typename T> size_t one_time_conveyor_feeder<T>::queued() const {
	if (feedee) {
		return feedee->queued();
	}
	return 0;
}

template <typename T> size_t one_time_conveyor_feeder<T>::space() const {
	if (feedee) {
		return feedee->space();
	}
	return 0;
}

template <typename T> OneTimeConveyorNode<T>::~OneTimeConveyorNode() {
	if (feeder) {
		feeder->setFeedee(nullptr);
		feeder = nullptr;
	}
}

template <typename T>
void OneTimeConveyorNode<T>::set_feeder(one_time_conveyor_feeder<T> *feeder_p) {
	feeder = feeder_p;
}

template <typename T> void OneTimeConveyorNode<T>::feed(T &&value) {
	storage = std::move(value);
	arm_next();
}

template <typename T> void OneTimeConveyorNode<T>::fail(Error &&error) {
	storage = std::move(error);
	arm_next();
}

template <typename T> size_t OneTimeConveyorNode<T>::queued() const {
	return storage.has_value() ? 1 : 0;
}

template <typename T> size_t OneTimeConveyorNode<T>::space() const {
	return passed ? 0 : 1;
}

template <typename T>
void OneTimeConveyorNode<T>::get_result(ErrorOrValue &err_or_val) {
	if (storage.has_value()) {
		err_or_val.as<T>() = std::move(storage.value());
		storage = std::nullopt;
	} else {
		err_or_val.as<T>() =
			criticalError("Signal for retrieval of storage sent even though no "
						  "data is present");
	}
}

template <typename T> void OneTimeConveyorNode<T>::fire() {
	if (parent) {
		parent->child_has_fired();
	}
}

} // namespace saw
