#pragma once

#include <string>
#include <string_view>
#include <variant>

#include <cassert>

#include "common.h"

namespace saw {
/**
 * Utility class for generating errors. Has a base distinction between
 * critical and recoverable errors. Additional code ids can be provided to the
 * constructor if additional distinctions are necessary.
 */
class error {
public:
	enum class Code : int16_t {
		GenericCritical = -1,
		GenericRecoverable = 1,
		Disconnected = -99,
		Exhausted = -98
	};

private:
	std::variant<std::string_view, std::string> m_error_message;
	Code m_error;

public:
	error();
	error(const std::string_view &msg, error::Code code);
	error(std::string &&msg, error::Code code);
	error(error &&error);

	SAW_FORBID_COPY(error);

	error &operator=(error &&) = default;

	const std::string_view message() const;
	bool failed() const;

	bool is_critical() const;
	bool is_recoverable() const;

	error copy_error() const;

	Code error_code() const;
};

error make_error(const std::string_view &generic, error::Code c);

template <typename Formatter>
error make_error(const Formatter &formatter, error::Code code,
				 const std::string_view &generic) {
	try {
		std::string error_msg = formatter();
		return error{std::move(error_msg), code};
	} catch (std::bad_alloc &) {
		return error{generic, code};
	}
}

error critical_error(const std::string_view &generic,
					 error::Code c = error::Code::GenericCritical);

template <typename Formatter>
error critical_error(const Formatter &formatter,
					 const std::string_view &generic,
					 error::Code c = error::Code::GenericCritical) {
	return makeError(formatter, c, generic);
}

error recoverable_error(const std::string_view &generic,
						error::Code c = error::Code::GenericRecoverable);

template <typename Formatter>
error recoverable_error(const Formatter &formatter,
						const std::string_view &generic,
						error::Code c = error::Code::GenericRecoverable) {
	return makeError(formatter, c, generic);
}

error no_error();

/**
 * Exception alternative. Since I code without exceptions this class is
 * essentially a kind of exception replacement.
 */
template <typename T> class ErrorOr;

class error_or_value {
public:
	virtual ~error_or_value() = default;

	template <typename T> ErrorOr<UnfixVoid<T>> &as() {
		return static_cast<ErrorOr<UnfixVoid<T>> &>(*this);
	}

	template <typename T> const ErrorOr<UnfixVoid<T>> &as() const {
		return static_cast<const ErrorOr<UnfixVoid<T>> &>(*this);
	}
};

template <typename T> class ErrorOr final : public error_or_value {
private:
	std::variant<error, FixVoid<T>> m_value_or_error_data;

	static_assert(!std::is_same_v<T, Void>, "Don't use internal private types");

public:
	ErrorOr() = default;
	ErrorOr(const FixVoid<T> &value) : m_value_or_error_data{value} {}

	ErrorOr(FixVoid<T> &&value) : m_value_or_error_data{std::move(value)} {}

	ErrorOr(const error &error) : m_value_or_error_data{error} {}
	ErrorOr(error &&error) : m_value_or_error_data{std::move(error)} {}

	bool is_value() const {
		return std::holds_alternative<FixVoid<T>>(m_value_or_error_data);
	}

	bool is_error() const {
		return std::holds_alternative<error>(m_value_or_error_data);
	}

	class error &error() {
		return std::get<error>(m_value_or_error_data);
	}

	const class error &error() const { return std::get<error>(m_value_or_error_data); }

	FixVoid<T> &value() { return std::get<FixVoid<T>>(m_value_or_error_data); }

	const FixVoid<T> &value() const {
		return std::get<FixVoid<T>>(m_value_or_error_data);
	}
};

template <typename T> class ErrorOr<ErrorOr<T>> {
private:
	ErrorOr() = delete;
};

} // namespace saw
