#pragma once

#include "async.h"
#include "common.h"
#include "io_helpers.h"

#include <string>
#include <variant>

namespace saw {
/*
 * Input stream
 */
class InputStream {
public:
	virtual ~InputStream() = default;

	virtual ErrorOr<size_t> read(void *buffer, size_t length) = 0;

	virtual Conveyor<void> read_ready() = 0;

	virtual Conveyor<void> on_read_disconnected() = 0;
};

/*
 * Output stream
 */
class OutputStream {
public:
	virtual ~OutputStream() = default;

	virtual ErrorOr<size_t> write(const void *buffer, size_t length) = 0;

	virtual Conveyor<void> write_ready() = 0;
};

/*
 * Io stream
 */
class io_stream : public InputStream, public OutputStream {
public:
	virtual ~io_stream() = default;
};

class async_input_stream {
public:
	virtual ~async_input_stream() = default;

	virtual void read(void *buffer, size_t min_length, size_t max_length) = 0;

	virtual Conveyor<size_t> read_done() = 0;
	virtual Conveyor<void> on_read_disconnected() = 0;
};

class async_output_stream {
public:
	virtual ~async_output_stream() = default;

	virtual void write(const void *buffer, size_t length) = 0;

	virtual Conveyor<size_t> write_done() = 0;
};

class async_io_stream final : public async_input_stream,
							  public async_output_stream {
private:
	Own<io_stream> m_stream;

	ConveyorSink m_read_ready;
	ConveyorSink m_write_ready;
	ConveyorSink m_read_disconnected;

	ReadTaskAndStepHelper m_read_stepper;
	WriteTaskAndStepHelper m_write_stepper;

public:
	async_io_stream(Own<io_stream> str);

	SAW_FORBID_COPY(async_io_stream);
	SAW_FORBID_MOVE(async_io_stream);

	void read(void *buffer, size_t length, size_t max_length) override;

	Conveyor<size_t> read_done() override;

	Conveyor<void> on_read_disconnected() override;

	void write(const void *buffer, size_t length) override;

	Conveyor<size_t> write_done() override;
};

class server {
public:
	virtual ~server() = default;

	virtual Conveyor<Own<io_stream>> accept() = 0;
};

class NetworkAddress;
/**
 * Datagram class. Bound to a local address it is able to receive inbound
 * datagram messages and send them as well as long as an address is provided as
 * well
 */
class datagram {
public:
	virtual ~datagram() = default;

	virtual ErrorOr<size_t> read(void *buffer, size_t length) = 0;
	virtual Conveyor<void> read_ready() = 0;

	virtual ErrorOr<size_t> write(const void *buffer, size_t length,
								  NetworkAddress &dest) = 0;
	virtual Conveyor<void> write_ready() = 0;
};

class OsNetworkAddress;
class StringNetworkAddress;

class NetworkAddress {
public:
	using ChildVariant =
		std::variant<OsNetworkAddress *, StringNetworkAddress *>;

	virtual ~NetworkAddress() = default;

	virtual NetworkAddress::ChildVariant representation() = 0;

	virtual const std::string &address() const = 0;
	virtual uint16_t port() const = 0;
};

class OsNetworkAddress : public NetworkAddress {
public:
	virtual ~OsNetworkAddress() = default;

	NetworkAddress::ChildVariant representation() override { return this; }
};

class StringNetworkAddress final : public NetworkAddress {
private:
	std::string m_address_value;
	uint16_t m_port_value;

public:
	StringNetworkAddress(const std::string &address, uint16_t port);

	const std::string &address() const override;
	uint16_t port() const override;

	NetworkAddress::ChildVariant representation() override { return this; }
};

class network {
public:
	virtual ~network() = default;

	/**
	 * Resolve the provided string and uint16 to the preferred storage method
	 */
	virtual Conveyor<Own<NetworkAddress>>
	resolve_address(const std::string &addr, uint16_t port_hint = 0) = 0;

	/**
	 * Parse the provided string and uint16 to the preferred storage method
	 * Since no dns request is made here, no async conveyors have to be used.
	 */
	/// @todo implement
	// virtual Own<NetworkAddress> parseAddress(const std::string& addr,
	// uint16_t port_hint = 0) = 0;

	/**
	 * Set up a listener on this address
	 */
	virtual Own<server> listen(NetworkAddress &bind_addr) = 0;

	/**
	 * Connect to a remote address
	 */
	virtual Conveyor<Own<io_stream>> connect(NetworkAddress &address) = 0;

	/**
	 * Bind a datagram socket at this address.
	 */
	virtual Own<datagram> datagram(NetworkAddress &address) = 0;
};

class io_provider {
public:
	virtual ~io_provider() = default;

	virtual Own<InputStream> wrap_input_fd(int fd) = 0;

	virtual network &network() = 0;
};

struct AsyncIoContext {
	Own<io_provider> io;
	EventLoop &event_loop;
	EventPort &event_port;
};

ErrorOr<AsyncIoContext> setup_async_io();
} // namespace saw
