#pragma once

#include "async.h"
#include "buffer.h"
#include "io.h"
#include "message.h"

namespace saw {

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer = MessageContainer<Incoming>,
		  typename OutContainer = MessageContainer<Outgoing>,
		  typename BufferT = RingBuffer>
class streaming_io_peer {
public:
	/**
	 *
	 */
	streaming_io_peer(
		Own<ConveyorFeeder<HeapMessageRoot<Incoming, InContainer>>> feed,
		Own<AsyncIoStream> stream, Codec codec, BufferT in, BufferT out);
	/**
	 *
	 */
	streaming_io_peer(
		Own<ConveyorFeeder<HeapMessageRoot<Incoming, InContainer>>> feed,
		Own<AsyncIoStream> stream);

	/**
	 * Deleted copy and move constructors
	 */
	SAW_FORBID_COPY(streaming_io_peer);
	SAW_FORBID_MOVE(streaming_io_peer);

	/**
	 * Send a message to the remote peer
	 */
	Error send(HeapMessageRoot<Outgoing, OutContainer> builder);

	/**
	 * A phantom conveyor feeder. Meant for interfacing with other components
	 */
	ConveyorFeeder<HeapMessageRoot<Outgoing, OutContainer>> &feeder();

	Conveyor<void> on_read_disconnected();

private:
	/// @unimplemented
	class peer_conveyor_feeder final
		: public ConveyorFeeder<HeapMessageRoot<Outgoing, OutContainer>> {
	public:
		peer_conveyor_feeder(
			streaming_io_peer<Codec, Incoming, Outgoing, InContainer,
							  OutContainer, BufferT> &peer)
			: m_peer{peer} {}

		void feed(HeapMessageRoot<Outgoing, OutContainer> &&data) override {
			(void)data;
		}

		void fail(Error &&error) override { (void)error; }

		size_t space() const override { return 0; }

		size_t queued() const override { return 0; }

	private:
		streaming_io_peer<Codec, Incoming, Outgoing, InContainer, OutContainer,
						  BufferT> &m_peer;
	};

private:
	Own<ConveyorFeeder<HeapMessageRoot<Incoming, InContainer>>>
		m_incoming_feeder = nullptr;

	Own<AsyncIoStream> m_io_stream;

	Codec m_codec;

	BufferT m_in_buffer;
	BufferT m_out_buffer;

	ConveyorSink m_sink_read;
	ConveyorSink m_sink_write;

	peer_conveyor_feeder m_conveyor_feeder;
};

/**
 * Setup new streaming io peer with the provided network protocols.
 * This is a convenience wrapper intended for a faster setup of this class
 */
template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer = MessageContainer<Incoming>,
		  typename OutContainer = MessageContainer<Outgoing>,
		  typename BufferT = RingBuffer>
std::pair<Own<streaming_io_peer<Codec, Incoming, Outgoing, InContainer,
								OutContainer, BufferT>>,
		  Conveyor<HeapMessageRoot<Incoming, InContainer>>>
new_streaming_io_peer(Own<AsyncIoStream> stream);

} // namespace saw

#include "io_peer.tmpl.h"
