namespace saw {

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
streaming_io_peer<Codec, Incoming, Outgoing, InContainer, OutContainer,
				  BufferT>::
	streaming_io_peer(
		Own<ConveyorFeeder<HeapMessageRoot<Incoming, InContainer>>> feed,
		Own<AsyncIoStream> str)
	: streaming_io_peer{std::move(feed), std::move(str), {}, {}, {}} {}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
streaming_io_peer<Codec, Incoming, Outgoing, InContainer, OutContainer,
				  BufferT>::
	streaming_io_peer(
		Own<ConveyorFeeder<HeapMessageRoot<Incoming, InContainer>>> feed_,
		Own<AsyncIoStream> stream_, Codec codec_, BufferT in_, BufferT out_)
	: incoming_feeder{std::move(feed_)}, io_stream{std::move(stream_)},
	  m_codec{std::move(codec_)}, m_in_buffer{std::move(in_)}, m_out_buffer{std::move(
															   out_)},
	  sink_read{io_stream->readDone()
					.then([streaming_io_peer](size_t bytes) -> ErrorOr<void> {
						in_buffer.writeAdvance(bytes);

						if (in_buffer.writeSegmentLength() == 0) {
							return criticalError("Message too long");
						}

						io_stream->read(&in_buffer.write(), 1,
										in_buffer.writeSegmentLength());

						while (true) {
							auto root =
								heapMessageRoot<Incoming, InContainer>();
							auto builder = root.build();

							Error error =
								codec.template decode<Incoming, InContainer>(
									builder, in_buffer);
							if (error.isCritical()) {
								return error;
							}

							if (!error.failed()) {
								incoming_feeder->feed(std::move(root));
							} else {
								break;
							}
						}

						return Void{};
					})
					.sink([streaming_io_peer](Error error) {
						incoming_feeder->fail(error.copyError());

						return error;
					})},
	  sink_write{io_stream->writeDone()
					 .then([streaming_io_peer](size_t bytes) -> ErrorOr<void> {
						 out_buffer.readAdvance(bytes);
						 if (out_buffer.readCompositeLength() > 0) {
							 io_stream->write(&out_buffer.read(),
											  out_buffer.readSegmentLength());
						 }

						 return Void{};
					 })
					 .sink()} {
	io_stream->read(&in_buffer.write(), 1, in_buffer.writeSegmentLength());
}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
Error streaming_io_peer<Codec, Incoming, Outgoing, InContainer, OutContainer,
						BufferT>::send(HeapMessageRoot<Outgoing, OutContainer>
										   msg) {
	bool restart_write = m_out_buffer.readSegmentLength() == 0;

	Error error =
		codec.template encode<Outgoing, OutContainer>(msg.read(), out_buffer);
	if (error.failed()) {
		return error;
	}

	if (restart_write) {
		io_stream->write(&out_buffer.read(), out_buffer.readSegmentLength());
	}

	return noError();
}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
Conveyor<void>
streaming_io_peer<Codec, Incoming, Outgoing, InContainer, OutContainer,
				  BufferT>::on_read_disconnected() {
	return io_stream->onReadDisconnected();
}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
std::pair<Own<streaming_io_peer<Codec, Incoming, Outgoing, InContainer,
								OutContainer, BufferT>>,
		  Conveyor<HeapMessageRoot<Incoming, InContainer>>>
new_streaming_io_peer(Own<AsyncIoStream> stream) {
	auto caf = newConveyorAndFeeder<HeapMessageRoot<Incoming, InContainer>>();

	return {heap<streaming_io_peer<Codec, Incoming, Outgoing, InContainer,
								   OutContainer, BufferT>>(
				std::move(caf.feeder), std::move(stream)),
			std::move(caf.conveyor)};
}

} // namespace saw
